==> intel-oneapi-mkl-2022.1.0_223-3-aarch64.pkg.tar.xz: c_offload/sparse_blas/source/sparse_csr_d_trsv.c <==
/*******************************************************************************
* Copyright 2021-2022 Intel Corporation.
*
* This software and the related documents are Intel copyrighted  materials,  and
* your use of  them is  governed by the  express license  under which  they were
* provided to you (License).  Unless the License provides otherwise, you may not
* use, modify, copy, publish, distribute,  disclose or transmit this software or
* the related documents without Intel's prior written permission.
*
* This software and the related documents  are provided as  is,  with no express
* or implied  warranties,  other  than those  that are  expressly stated  in the
* License.
*******************************************************************************/

/*
*   Content : Intel(R) oneAPI Math Kernel Library (oneMKL) Sparse BLAS C OpenMP
*             offload example for mkl_sparse_d_trsv
*
********************************************************************************
*
* Consider the matrix A (see 'Sparse Storage Formats for Sparse BLAS Level 2
* and Level 3 in the  Intel oneMKL Reference Manual')
*
*                 |   1       -1      0   -3     0   |
*                 |  -2        5      0    0     0   |
*   A    =        |   0        0      4    6     4   |,
*                 |  -4        0      2    7     0   |
*                 |   0        8      0    0    -5   |
*
*  The matrix A is represented in a one-based compressed sparse row (CSR) storage
*  scheme with three arrays (see 'Sparse Matrix Storage Schemes' in the
*   Intel oneMKL Reference Manual) as follows:
*
*         values  =  ( 1 -1 -3 -2  5  4  6  4 -4  2  7  8 -5 )
*         columns =  ( 1  2  4  1  2  3  4  5  1  3  4  2  5 )
*         rowIndex = ( 1        4     6        9       12    14 )
*
*  The test performs the operation
*
*       L^{-1}*x = y, using mkl_sparse_d_trsv
*
*  Here A is a general sparse matrix
*        L is the lower triangular part of matrix A with non-unit diagonal
*        x, y and z are vectors
*
********************************************************************************
*/
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "common_for_sparse_examples.h"
#include "mkl.h"
#include "mkl_omp_offload.h"

int main() {
    //*******************************************************************************
    //     Declaration and initialization of parameters for sparse representation of
    //     the matrix A in the compressed sparse row format:
    //*******************************************************************************
#define M 5
#define N 5
#define NNZ 13
    // Descriptor of main sparse matrix properties
    struct matrix_descr descrA;

    // // Structure with sparse matrix stored in CSR format
    sparse_matrix_t csrA;
    //*******************************************************************************
    //    Sparse representation of the matrix A
    //*******************************************************************************
    double alpha = 1.0;
    MKL_INT i;

    double *values     = (double *)mkl_malloc(sizeof(double) * NNZ, 64);
    MKL_INT *columns   = (MKL_INT *)mkl_malloc(sizeof(MKL_INT) * NNZ, 64);
    MKL_INT *row_index = (MKL_INT *)mkl_malloc(sizeof(MKL_INT) * (N + 1), 64);
    double *x          = (double *)mkl_malloc(sizeof(double) * M, 64);

    double *y = (double *)mkl_malloc(sizeof(double) * N, 64);
    double *z = (double *)mkl_malloc(sizeof(double) * N, 64);

    const int num_pointers = 6;
    void *pointer_array[num_pointers];
    pointer_array[0] = values;
    pointer_array[1] = columns;
    pointer_array[2] = row_index;
    pointer_array[3] = x;
    pointer_array[4] = y;
    pointer_array[5] = z;

    if (!values || !columns || !row_index || !x || !y || !z) {
        free_allocated_memories(pointer_array, num_pointers);
        return 1;
    }

    //*******************************************************************************
    //    Sparse representation of the matrix A
    //*******************************************************************************
    double init_values[NNZ] = {1.0, -1.0, -3.0, -2.0, 5.0, 4.0, 6.0,
                               4.0, -4.0, 2.0,  7.0,  8.0, -5.0};
    MKL_INT init_columns[NNZ]     = {1, 2, 4, 1, 2, 3, 4, 5, 1, 3, 4, 2, 5};
    MKL_INT init_row_index[N + 1] = {1, 4, 6, 9, 12, 14};
    double init_x[N]              = {1.0, 5.0, 1.0, 4.0, 1.0};

    for (i = 0; i < NNZ; i++) {
        values[i]  = init_values[i];
        columns[i] = init_columns[i];
    }
    for (i = 0; i < N + 1; i++) {
        row_index[i] = init_row_index[i];
    }

    for (i   = 0; i < N; i++)
        x[i] = init_x[i];
    for (i = 0; i < M; i++) {
        y[i] = 0.0;
        z[i] = 0.0;
    }

    printf("\n EXAMPLE PROGRAM FOR mkl_sparse_d_trsv omp_offload \n");
    printf("-----------------------------------------------------\n");
    printf("\n" );
    printf("   INPUT DATA FOR mkl_sparse_d_trsv omp offload   \n");
    printf("   LOWER TRIANGULAR SPARSE MATRIX \n");
    printf("   WITH NON UNIT DIAGONAL \n");
    printf("   ALPHA = %4.1f \n", alpha);
    printf("   SPARSE_OPERATION_NON_TRANSPOSE \n");
    printf("   Input vector                   \n");
    for ( i = 0; i < N; i++ ) {
        printf("%7.1f\n", x[i]);
    }

    // Create matrix descriptor
    descrA.type = SPARSE_MATRIX_TYPE_TRIANGULAR;
    descrA.mode = SPARSE_FILL_MODE_LOWER;
    descrA.diag = SPARSE_DIAG_NON_UNIT;

    sparse_status_t ie_status;

    // Create handle with matrix stored in CSR format
    ie_status = mkl_sparse_d_create_csr(&csrA, SPARSE_INDEX_BASE_ONE,
                                        N, // number of rows
                                        M, // number of cols
                                        row_index, row_index + 1, columns, values);
    if (ie_status != SPARSE_STATUS_SUCCESS) {
        printf(" Error in mkl_sparse_d_create_csr: %d\n", ie_status);
        free_allocated_memories(pointer_array, num_pointers);
        return ie_status;
    }

    ie_status = mkl_sparse_set_sv_hint(csrA, SPARSE_OPERATION_NON_TRANSPOSE, descrA, 1);
    if (ie_status != SPARSE_STATUS_SUCCESS && ie_status != SPARSE_STATUS_NOT_SUPPORTED) {
        printf(" Error in set_sv_hint: %d\n", ie_status);
        free_allocated_memories(pointer_array, num_pointers);
        return ie_status;
    }

    ie_status = mkl_sparse_optimize(csrA);
    if (ie_status != SPARSE_STATUS_SUCCESS) {
        printf(" Error in mkl_sparse_optimize: %d\n", ie_status);
        free_allocated_memories(pointer_array, num_pointers);
        return ie_status;
    }

    // Compute y = alpha * A^{-1} * x
    ie_status = mkl_sparse_d_trsv(SPARSE_OPERATION_NON_TRANSPOSE, alpha, csrA, descrA, x, y);
    if (ie_status != SPARSE_STATUS_SUCCESS) {
        printf(" Error in mkl_sparse_d_trsv: %d\n", ie_status);
        free_allocated_memories(pointer_array, num_pointers);
        return ie_status;
    }

    // Release matrix handle and deallocate matrix
    ie_status = mkl_sparse_destroy(csrA);
    if (ie_status != SPARSE_STATUS_SUCCESS) {
        printf(" Error in mkl_sparse_destroy: %d\n", ie_status);
        free_allocated_memories(pointer_array, num_pointers);
        return ie_status;
    }

    printf("\n");
    printf("   OUTPUT DATA FOR mkl_sparse_d_trsv \n");

    // y should be equal { 1.0000, 1.4000, 0.2500, 1.0714, 2.0400 }
    for ( i = 0; i < N; i++ ) {
        printf("%7.4f\n", y[i]);
    }
    printf("---------------------------------------------------\n");
    fflush(stdout);

    const int devNum = 0;

    sparse_matrix_t csrA_gpu;

    sparse_status_t status_create;
    sparse_status_t status_hint;
    sparse_status_t status_opt;
    sparse_status_t status_trsv;
    sparse_status_t status_destroy;

// call create_csr/set_sv_hint/optimize/trsv/destroy via omp_offload.
#pragma omp target data map(to:row_index[0:N+1],columns[0:NNZ],values[0:NNZ],x[0:M])          \
                        map(tofrom:z[0:N]) device(devNum)
    {
        printf("Create CSR matrix via omp_offload\n");

#pragma omp target variant dispatch device(devNum) use_device_ptr(row_index, columns, values)
        status_create = mkl_sparse_d_create_csr(&csrA_gpu, SPARSE_INDEX_BASE_ONE, N, M,
                                                row_index, row_index + 1, columns, values);

        printf("mkl_sparse_set_sv_hint() ... \n");
#pragma omp target variant dispatch device(devNum)
        status_hint = mkl_sparse_set_sv_hint(csrA_gpu, SPARSE_OPERATION_NON_TRANSPOSE, descrA, 1);

        printf("mkl_sparse_optimize() ... \n");
#pragma omp target variant dispatch device(devNum)
        status_opt = mkl_sparse_optimize(csrA_gpu);

        printf("Compute mkl_sparse_d_trsv via omp_offload\n");

#pragma omp target variant dispatch device(devNum) use_device_ptr(x, z)
        status_trsv = mkl_sparse_d_trsv(SPARSE_OPERATION_NON_TRANSPOSE, alpha, csrA_gpu, descrA, x, z);

        printf("Destroy the CSR matrix via omp_offload\n");

#pragma omp target variant dispatch device(devNum)
        status_destroy = mkl_sparse_destroy(csrA_gpu);

        printf("End of mkl_sparse_destroy omp offload call.\n");
    }
    printf("End all offload function calls.\n"); fflush(stdout);

    int flps_per_value = 2 * ((NNZ / N) + 1);
    int status         = 0;

    int status_offload = status_create | status_hint | status_opt | status_trsv | status_destroy;
    if (status_offload != 0) {
        printf("\tERROR: status_create = %d, status_hint = %d, status_opt = %d, "
               "status_trsv = %d, status_destroy = %d\n",
               status_create, status_hint, status_opt, status_trsv, status_destroy);
        fflush(stdout);

        goto cleanup;
    }

    printf("   OUTPUT DATA FOR mkl_sparse_d_trsv_omp_offload \n");
    // z should be equal { 1.0000, 1.4000, 0.2500, 1.0714, 2.0400 }
    for (i = 0; i < N; i++) {
        printf("%7.4f\n", z[i]);
    }
    printf("---------------------------------------------------\n");
    fflush(stdout);

    status = validation_result_double(y, z, M, flps_per_value);

cleanup:
    free_allocated_memories(pointer_array, num_pointers);

    const int status_all = status | status_offload;
    printf("Test %s\n", status_all == 0 ? "PASSED" : "FAILED");
    fflush(stdout);

    return status_all;
}

==> intel-oneapi-mkl-2023.2.0_49495-2-aarch64.pkg.tar.xz: c_offload/sparse_blas/source/sparse_csr_d_trsv.c <==
/*******************************************************************************
* Copyright 2021-2022 Intel Corporation.
*
* This software and the related documents are Intel copyrighted  materials,  and
* your use of  them is  governed by the  express license  under which  they were
* provided to you (License).  Unless the License provides otherwise, you may not
* use, modify, copy, publish, distribute,  disclose or transmit this software or
* the related documents without Intel's prior written permission.
*
* This software and the related documents  are provided as  is,  with no express
* or implied  warranties,  other  than those  that are  expressly stated  in the
* License.
*******************************************************************************/

/*
*   Content : Intel(R) oneAPI Math Kernel Library (oneMKL) Sparse BLAS C OpenMP
*             offload example for mkl_sparse_d_trsv
*
********************************************************************************
*
* Consider the matrix A (see 'Sparse Storage Formats for Sparse BLAS Level 2
* and Level 3 in the  Intel oneMKL Reference Manual')
*
*                 |   1       -1      0   -3     0   |
*                 |  -2        5      0    0     0   |
*   A    =        |   0        0      4    6     4   |,
*                 |  -4        0      2    7     0   |
*                 |   0        8      0    0    -5   |
*
*  The matrix A is represented in a one-based compressed sparse row (CSR) storage
*  scheme with three arrays (see 'Sparse Matrix Storage Schemes' in the
*   Intel oneMKL Reference Manual) as follows:
*
*         values  =  ( 1 -1 -3 -2  5  4  6  4 -4  2  7  8 -5 )
*         columns =  ( 1  2  4  1  2  3  4  5  1  3  4  2  5 )
*         rowIndex = ( 1        4     6        9       12    14 )
*
*  The test performs the operation
*
*       L^{-1}*x = y, using mkl_sparse_d_trsv
*
*  Here A is a general sparse matrix
*        L is the lower triangular part of matrix A with non-unit diagonal
*        x, y and z are vectors
*
********************************************************************************
*/
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "common_for_sparse_examples.h"
#include "mkl.h"
#include "mkl_omp_offload.h"

int main() {
    //*******************************************************************************
    //     Declaration and initialization of parameters for sparse representation of
    //     the matrix A in the compressed sparse row format:
    //*******************************************************************************
#define M 5
#define N 5
#define NNZ 13
    // Descriptor of main sparse matrix properties
    struct matrix_descr descrA;

    // // Structure with sparse matrix stored in CSR format
    sparse_matrix_t csrA;
    //*******************************************************************************
    //    Sparse representation of the matrix A
    //*******************************************************************************
    double alpha = 1.0;
    MKL_INT i;

    double *values     = (double *)mkl_malloc(sizeof(double) * NNZ, 64);
    MKL_INT *columns   = (MKL_INT *)mkl_malloc(sizeof(MKL_INT) * NNZ, 64);
    MKL_INT *row_index = (MKL_INT *)mkl_malloc(sizeof(MKL_INT) * (N + 1), 64);
    double *x          = (double *)mkl_malloc(sizeof(double) * M, 64);

    double *y = (double *)mkl_malloc(sizeof(double) * N, 64);
    double *z = (double *)mkl_malloc(sizeof(double) * N, 64);

    const int num_pointers = 6;
    void *pointer_array[num_pointers];
    pointer_array[0] = values;
    pointer_array[1] = columns;
    pointer_array[2] = row_index;
    pointer_array[3] = x;
    pointer_array[4] = y;
    pointer_array[5] = z;

    if (!values || !columns || !row_index || !x || !y || !z) {
        free_allocated_memories(pointer_array, num_pointers);
        return 1;
    }

    //*******************************************************************************
    //    Sparse representation of the matrix A
    //*******************************************************************************
    double init_values[NNZ] = {1.0, -1.0, -3.0, -2.0, 5.0, 4.0, 6.0,
                               4.0, -4.0, 2.0,  7.0,  8.0, -5.0};
    MKL_INT init_columns[NNZ]     = {1, 2, 4, 1, 2, 3, 4, 5, 1, 3, 4, 2, 5};
    MKL_INT init_row_index[N + 1] = {1, 4, 6, 9, 12, 14};
    double init_x[N]              = {1.0, 5.0, 1.0, 4.0, 1.0};

    for (i = 0; i < NNZ; i++) {
        values[i]  = init_values[i];
        columns[i] = init_columns[i];
    }
    for (i = 0; i < N + 1; i++) {
        row_index[i] = init_row_index[i];
    }

    for (i   = 0; i < N; i++)
        x[i] = init_x[i];
    for (i = 0; i < M; i++) {
        y[i] = 0.0;
        z[i] = 0.0;
    }

    printf("\n EXAMPLE PROGRAM FOR mkl_sparse_d_trsv omp_offload \n");
    printf("-----------------------------------------------------\n");
    printf("\n" );
    printf("   INPUT DATA FOR mkl_sparse_d_trsv omp offload   \n");
    printf("   LOWER TRIANGULAR SPARSE MATRIX \n");
    printf("   WITH NON UNIT DIAGONAL \n");
    printf("   ALPHA = %4.1f \n", alpha);
    printf("   SPARSE_OPERATION_NON_TRANSPOSE \n");
    printf("   Input vector                   \n");
    for ( i = 0; i < N; i++ ) {
        printf("%7.1f\n", x[i]);
    }

    // Create matrix descriptor
    descrA.type = SPARSE_MATRIX_TYPE_TRIANGULAR;
    descrA.mode = SPARSE_FILL_MODE_LOWER;
    descrA.diag = SPARSE_DIAG_NON_UNIT;

    sparse_status_t ie_status;

    // Create handle with matrix stored in CSR format
    ie_status = mkl_sparse_d_create_csr(&csrA, SPARSE_INDEX_BASE_ONE,
                                        N, // number of rows
                                        M, // number of cols
                                        row_index, row_index + 1, columns, values);
    if (ie_status != SPARSE_STATUS_SUCCESS) {
        printf(" Error in mkl_sparse_d_create_csr: %d\n", ie_status);
        free_allocated_memories(pointer_array, num_pointers);
        return ie_status;
    }

    ie_status = mkl_sparse_set_sv_hint(csrA, SPARSE_OPERATION_NON_TRANSPOSE, descrA, 1);
    if (ie_status != SPARSE_STATUS_SUCCESS && ie_status != SPARSE_STATUS_NOT_SUPPORTED) {
        printf(" Error in set_sv_hint: %d\n", ie_status);
        free_allocated_memories(pointer_array, num_pointers);
        return ie_status;
    }

    ie_status = mkl_sparse_optimize(csrA);
    if (ie_status != SPARSE_STATUS_SUCCESS) {
        printf(" Error in mkl_sparse_optimize: %d\n", ie_status);
        free_allocated_memories(pointer_array, num_pointers);
        return ie_status;
    }

    // Compute y = alpha * A^{-1} * x
    ie_status = mkl_sparse_d_trsv(SPARSE_OPERATION_NON_TRANSPOSE, alpha, csrA, descrA, x, y);
    if (ie_status != SPARSE_STATUS_SUCCESS) {
        printf(" Error in mkl_sparse_d_trsv: %d\n", ie_status);
        free_allocated_memories(pointer_array, num_pointers);
        return ie_status;
    }

    // Release matrix handle and deallocate matrix
    ie_status = mkl_sparse_destroy(csrA);
    if (ie_status != SPARSE_STATUS_SUCCESS) {
        printf(" Error in mkl_sparse_destroy: %d\n", ie_status);
        free_allocated_memories(pointer_array, num_pointers);
        return ie_status;
    }

    printf("\n");
    printf("   OUTPUT DATA FOR mkl_sparse_d_trsv \n");

    // y should be equal { 1.0000, 1.4000, 0.2500, 1.0714, 2.0400 }
    for ( i = 0; i < N; i++ ) {
        printf("%7.4f\n", y[i]);
    }
    printf("---------------------------------------------------\n");
    fflush(stdout);

    const int devNum = 0;

#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
    printf("Using OpenMP 5.1. dispatch construct.\n"); fflush(0);
#else
    printf("Using OpenMP offload 'target variant dispatch' construct.\n"); fflush(0);
#endif

    sparse_matrix_t csrA_gpu;

    sparse_status_t status_create;
    sparse_status_t status_hint;
    sparse_status_t status_opt;
    sparse_status_t status_trsv;
    sparse_status_t status_destroy;

// call create_csr/set_sv_hint/optimize/trsv/destroy via omp_offload.
#pragma omp target data map(to:row_index[0:N+1],columns[0:NNZ],values[0:NNZ],x[0:M])          \
                        map(tofrom:z[0:N]) device(devNum)
    {
        printf("Create CSR matrix via omp_offload\n");

#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
#pragma omp dispatch device(devNum)
#else
#pragma omp target variant dispatch device(devNum) use_device_ptr(row_index, columns, values)
#endif
        status_create = mkl_sparse_d_create_csr(&csrA_gpu, SPARSE_INDEX_BASE_ONE, N, M,
                                                row_index, row_index + 1, columns, values);

        printf("mkl_sparse_set_sv_hint() ... \n");
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
#pragma omp dispatch device(devNum)
#else
#pragma omp target variant dispatch device(devNum)
#endif
        status_hint = mkl_sparse_set_sv_hint(csrA_gpu, SPARSE_OPERATION_NON_TRANSPOSE, descrA, 1);

        printf("mkl_sparse_optimize() ... \n");
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
#pragma omp dispatch device(devNum)
#else
#pragma omp target variant dispatch device(devNum)
#endif
        status_opt = mkl_sparse_optimize(csrA_gpu);

        printf("Compute mkl_sparse_d_trsv via omp_offload\n");

#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
#pragma omp dispatch device(devNum)
#else
#pragma omp target variant dispatch device(devNum) use_device_ptr(x, z)
#endif
        status_trsv = mkl_sparse_d_trsv(SPARSE_OPERATION_NON_TRANSPOSE, alpha, csrA_gpu, descrA, x, z);

        printf("Destroy the CSR matrix via omp_offload\n");

#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
#pragma omp dispatch device(devNum)
#else
#pragma omp target variant dispatch device(devNum)
#endif
        status_destroy = mkl_sparse_destroy(csrA_gpu);

        printf("End of mkl_sparse_destroy omp offload call.\n");
    }
    printf("End all offload function calls.\n"); fflush(stdout);

    int flps_per_value = 2 * ((NNZ / N) + 1);
    int status         = 0;

    int status_offload = status_create | status_hint | status_opt | status_trsv | status_destroy;
    if (status_offload != 0) {
        printf("\tERROR: status_create = %d, status_hint = %d, status_opt = %d, "
               "status_trsv = %d, status_destroy = %d\n",
               status_create, status_hint, status_opt, status_trsv, status_destroy);
        fflush(stdout);

        goto cleanup;
    }

    printf("   OUTPUT DATA FOR mkl_sparse_d_trsv_omp_offload \n");
    // z should be equal { 1.0000, 1.4000, 0.2500, 1.0714, 2.0400 }
    for (i = 0; i < N; i++) {
        printf("%7.4f\n", z[i]);
    }
    printf("---------------------------------------------------\n");
    fflush(stdout);

    status = validation_result_double(y, z, M, flps_per_value);

cleanup:
    free_allocated_memories(pointer_array, num_pointers);

    const int status_all = status | status_offload;
    printf("Test %s\n", status_all == 0 ? "PASSED" : "FAILED");
    fflush(stdout);

    return status_all;
}
